/*
 * uninformed research (c)
 * -------------------
 *
 * elf binary signature verification API
 *
 * skape
 * mmiller@hick.org
 * 01/29/2003
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "openssl/objects.h"
#include "openssl/rsa.h"
#include "openssl/md5.h"

#include "openssl/engine.h"
#include "openssl/x509.h"
#include "openssl/pem.h"
#include "openssl/bio.h"

#include "melf.h"
#include "verify.h"

typedef struct _elfverify_ctx {

	MELF *melf;

	unsigned char digest[16];

} ELFVERIFY_CTX;

X509_STORE *getX509Store(const char *caFile, const char *caPath);
unsigned long _elfverifyCallback(void *userParam, X509 *signer, unsigned long trustedIssuer);

unsigned long elfverifyTrust(const char *elfImage, const char *caFile, const char *caDirectory)
{
	return elfverifyTrustCallback(elfImage, caFile, caDirectory, _elfverifyCallback, NULL);
}

unsigned long elfverifyTrustCallback(const char *elfImage, const char *caFile, const char *caDirectory, unsigned long (*callback)(void *userParam, X509 *signer, unsigned long issuerTrusted), void *userParam)
{
	X509_STORE *store = getX509Store(caFile, caDirectory);
	STACK_OF(PKCS7_SIGNER_INFO) *signerStack = NULL;
	ELF_SPEC_HEADER *curr, *sigHeader = NULL;
	PKCS7_SIGNER_INFO *currSigner = NULL;
	unsigned long ret = 0, ok = 0, index = 0;
	BIO *sigRaw = NULL;
	ELFVERIFY_CTX ctx;
	PKCS7 *sig = NULL;
	MD5_CTX md5;

	memset(&ctx, 0, sizeof(ctx));

	MD5_Init(&md5);

	do
	{
		if (!(ctx.melf = melf_open(elfImage)))
			break;

		// Calculate binary checksum
		for (curr = melf_sectionGetEnum(ctx.melf);
				curr;
				curr = melf_sectionEnumNext(ctx.melf, curr))
		{
			const char *name = melf_sectionGetName(ctx.melf, curr);

			if (name && !strcmp(name, ".sig"))
			{
				// Two signatures?
				if (sigHeader)
					break;

				sigHeader = curr;

				continue;
			}

			if (curr->content)
				MD5_Update(&md5, curr->content, curr->contentLength);
		}

		MD5_Final(ctx.digest, &md5);

		// Not signed
		if (!sigHeader)
		{
			ret = ELFVERIFY_NOT_SIGNED;
			break;
		}

		// Extract PKCS7

		sigRaw = BIO_new_mem_buf(sigHeader->content, sigHeader->contentLength);
		sig    = PEM_read_bio_PKCS7(sigRaw, NULL, NULL, NULL);

		if (!sig)
		{
			ret = ELFVERIFY_INVALID_DIGEST;
			break;
		}

		signerStack = PKCS7_get_signer_info(sig);

		ok = 0;

		while ((signerStack) && (currSigner = sk_PKCS7_SIGNER_INFO_value(signerStack, index++)))
		{
			STACK_OF(X509) *certStack = sk_X509_new_null();
			X509 *cert = PKCS7_cert_from_signer_info(sig, currSigner);
			BIO  *outDigest = BIO_new(BIO_s_mem());
			unsigned char *outDigestMd5 = NULL;
			unsigned long outDigestMd5Length = 0, issuerTrusted = 0;

			if ((!certStack) || (!outDigest))
				break;

			sk_X509_push(certStack, cert);

			// Verify the actual md5 digest
			if (PKCS7_verify(sig, certStack, store, NULL, outDigest, PKCS7_TEXT) <= 0)
				issuerTrusted = 0;
			else
				issuerTrusted = 1;

			if (callback)
				ok = callback(userParam, cert, issuerTrusted);

			if (!ok)
				break;

			// Calculate the digest
			outDigestMd5Length = BIO_get_mem_data(outDigest, &outDigestMd5);

			if (outDigestMd5Length == sizeof(ctx.digest))
			{
				// Digest's mismatch, binary was modified.
				if (memcmp(ctx.digest, outDigestMd5, outDigestMd5Length))
				{
					ret = ELFVERIFY_INVALID_DIGEST;
					ok  = 0;
				}
			}

			if (certStack)
				sk_X509_free(certStack);
			if (outDigest)
				BIO_free(outDigest);

			if (!ok)
				break;
		}

		if (ret == 0 && !ok)
		{
			ret = ELFVERIFY_CERT_UNTRUSTED;
			break;
		}
		else if (!ok)
			break;

		ret = ELFVERIFY_SUCCESS;

	} while (0);

	if (store)
		X509_STORE_free(store);
	if (sigRaw)
		BIO_free(sigRaw);
	if (sig)
		PKCS7_free(sig);
	if (ctx.melf)
		melf_destroy(ctx.melf);

	return ret;
}

unsigned long _elfverifyCallback(void *userParam, X509 *signer, unsigned long trustedIssuer)
{
	return (trustedIssuer) ? 1 : 0;
}

X509_STORE *getX509Store(const char *caFile, const char *caPath)
{
	X509_STORE *store = NULL;
	X509_LOOKUP *look = NULL;

	do
	{
		if (!(store = X509_STORE_new()))
			break;

		look = X509_STORE_add_lookup(store, X509_LOOKUP_file());

		if (look && caFile)
		{
			if (!X509_LOOKUP_load_file(look, caFile, X509_FILETYPE_PEM))
				break;
		}
		else if (look)
			X509_LOOKUP_load_file(look, NULL, X509_FILETYPE_DEFAULT);

		look = X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir());

		if (look && caPath)
		{
			if (!X509_LOOKUP_add_dir(look, caPath, X509_FILETYPE_PEM))
				break;
		}
		else if (look)
			X509_LOOKUP_add_dir(look, NULL, X509_FILETYPE_DEFAULT);

	} while (0);

	return store;
}

