/*
 * uninformed research (c)
 * -------------------
 *
 * elf binary signing tool.
 *
 * skape
 * mmiller@hick.org
 * 01/29/2003
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "openssl/objects.h"
#include "openssl/rsa.h"
#include "openssl/md5.h"

#include "openssl/engine.h"
#include "openssl/x509.h"
#include "openssl/pem.h"
#include "openssl/bio.h"

#include "melf.h"

#include "../elfsign.h"

typedef struct _elfsign_ctx {

	MELF *melf;

	char digest[16];

	// Files
	const char *elfFile;
	const char *certificateFile;
	const char *privateKeyFile;

	// OpenSSL
	
	X509     *certificate;
	EVP_PKEY *privateKey;

} ELFSIGN_CTX;

unsigned long elfsignImageOpen(ELFSIGN_CTX *ctx);
unsigned long elfsignImageSetSigSection(ELFSIGN_CTX *ctx);
unsigned long elfsignImageSave(ELFSIGN_CTX *ctx, const char *path);
unsigned long elfsignImageClose(ELFSIGN_CTX *ctx);

X509     *loadCertificate(const char *path);
EVP_PKEY *loadKey(const char *path);
int      passwordPrompt(char *buf, int bufSize, int verify, void *str);

int main(int argc, char **argv)
{
	extern char *optarg;
	ELFSIGN_CTX ctx;
	int c;

	// Initialize OpenSSL
   CRYPTO_malloc_init();
   ERR_load_crypto_strings();
   OpenSSL_add_all_algorithms();
   ENGINE_load_builtin_engines();

	memset(&ctx, 0, sizeof(ctx));

	while ((c = getopt(argc, argv, "f:c:p:hv")) != EOF)
	{
		switch (c)
		{
			case 'f':
				ctx.elfFile = optarg;
				break;
			case 'c':
				ctx.certificateFile = optarg;
				break;
			case 'p':
				ctx.privateKeyFile = optarg;
				break;
			case 'h':
				fprintf(stdout, "\n"
									 "elfsign -- sign elf images\n"
									 "\n"
									 "Usage: ./elfsign [-f elf image] [-c certificate file] [-p private key file] [-h]\n"
									 "\n");
				return 0;
			case 'v':
				fprintf(stdout, "elfsign %s (c) Uninformed Research\n"
									 "               skape - mmiller@hick.org\n",
									 ELFSIGN_VERSION);
				return 0;
			default:
				break;
		}
	}

	do
	{
		if (!ctx.elfFile)
		{
			fprintf(stdout, "error: no elf image specified.\n");
			break;
		}
		
		if (!ctx.certificateFile)
		{
			fprintf(stdout, "error: no certificate specified.\n");
			break;
		}
		
		if (!ctx.privateKeyFile)
		{
			fprintf(stdout, "error: no private key specified.\n");
			break;
		}

		// Now for the real deal

		if (!elfsignImageOpen(&ctx))
		{
			fprintf(stdout, "error: could not open elf image '%s'.", ctx.elfFile);
			break;
		}

		if (!elfsignImageSetSigSection(&ctx))
		{
			fprintf(stdout, "error: failed to sign elf image '%s'.", ctx.elfFile);
			break;
		}

		if (!elfsignImageSave(&ctx, ctx.elfFile))
		{
			fprintf(stdout, "error: failed to save elf image '%s'.", ctx.elfFile);
			break;
		}

		elfsignImageClose(&ctx);

	} while (0);

	return 1;
}

unsigned long elfsignImageOpen(ELFSIGN_CTX *ctx)
{
	ctx->melf = melf_open(ctx->elfFile);

	if (!ctx->melf)
		return 0;

	// Load the certificate
	ctx->certificate = loadCertificate(ctx->certificateFile);

	// Load the private key
	ctx->privateKey  = loadKey(ctx->privateKeyFile);

	if (!ctx->certificate || !ctx->privateKey)
		return 0;

	return 1;
}

unsigned long elfsignImageSetSigSection(ELFSIGN_CTX *ctx)
{
	ELF_SPEC_HEADER *curr = NULL, *sigHeader = NULL;
	unsigned long ret = 0, bufLength = 0;
	BIO *sigRaw = BIO_new(BIO_s_mem()), *digestRaw = NULL;
	unsigned char *buf = NULL;
	PKCS7 *sig7 = NULL;
	MD5_CTX md5;

	do
	{
		MD5_Init(&md5);

		sigHeader = melf_sectionFindName(ctx->melf, ".sig");
	
		// Create the signature section before doing the checksums as we potentially modify the string table for sections.
		if (!sigHeader)
			sigHeader = melf_sectionAdd(ctx->melf);
	
		if (!sigHeader)
			break;
	
		melf_sectionSetType(ctx->melf, sigHeader, SHT_PROGBITS);
		melf_sectionSetAddress(ctx->melf, sigHeader, 0);
		melf_sectionSetFlags(ctx->melf, sigHeader, 0);
		melf_sectionSetName(ctx->melf, sigHeader, ".sig");
		melf_sectionSetContent(ctx->melf, sigHeader, NULL, 0);
		
		for (curr = melf_sectionGetEnum(ctx->melf);
				curr;
				curr = melf_sectionEnumNext(ctx->melf, curr))
		{
			const char *name = melf_sectionGetName(ctx->melf, curr);

			// Don't checksum the sig header.
			if (name && !strcmp(name, ".sig"))
				continue;
	
			if (curr->content)
				MD5_Update(&md5, curr->content, curr->contentLength);
		}
	
		MD5_Final(ctx->digest, &md5);

		if (!(digestRaw = BIO_new_mem_buf(ctx->digest, sizeof(ctx->digest))))
			break;
	
		sig7 = PKCS7_sign(ctx->certificate, ctx->privateKey, NULL, digestRaw, PKCS7_TEXT);

		if (!sig7)
			break;

		if (!PEM_write_bio_PKCS7(sigRaw, sig7))
			break;

		bufLength = BIO_get_mem_data(sigRaw, &buf);

		melf_sectionSetContent(ctx->melf, sigHeader, buf, bufLength);

		ret = 1;

	} while (0);

	if (sigRaw)
		BIO_free(sigRaw);
	if (digestRaw)
		BIO_free(digestRaw);
	if (sig7)
		PKCS7_free(sig7);

	return ret;
}

unsigned long elfsignImageSave(ELFSIGN_CTX *ctx, const char *path)
{
	melf_save(ctx->melf, path);

	return 1;
}

unsigned long elfsignImageClose(ELFSIGN_CTX *ctx)
{
	if (ctx->melf)
		melf_destroy(ctx->melf);

	ctx->melf = NULL;

	return 1;
}

X509 *loadCertificate(const char *path)
{
	X509 *ret = NULL;
	BIO *bio  = BIO_new_file(path, "r");

	do
	{
		if (!bio)
			break;
	
		ret = PEM_read_bio_X509_AUX(bio, NULL, passwordPrompt, "Certificate Password: ");
	
	} while (0);

	if (bio)
		BIO_free(bio);

	return ret;
}

EVP_PKEY *loadKey(const char *path)
{
	EVP_PKEY *ret = NULL;
	BIO *bio      = BIO_new(BIO_s_file());

	do
	{
		if (!bio)
			break;

		if (BIO_read_filename(bio, path) <= 0)
			break;

		ret = PEM_read_bio_PrivateKey(bio, NULL, passwordPrompt, "Key Password: ");

	} while (0);

	if (bio)
		BIO_free(bio);
	
	return ret;
}

int passwordPrompt(char *buf, int bufSize, int verify, void *str)
{
	char *ret = getpass((char *)str);

	if (ret)
		strncpy(buf, ret, bufSize);

	return strlen(buf);
}
